@@ -1,6 +1,6 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009, 2010, 2011  Red Hat, Inc
+ * Copyright (C) 2002, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013  Red Hat, Inc
  * Author: Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  */
 
 #include <stdarg.h>
@@ -37,6 +37,25 @@
 static pthread_key_t tls_key = -1;
 static pthread_once_t tls_key_once = PTHREAD_ONCE_INIT;
 
+#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
+# define CUPS_PRINTER_DISCOVERED	0x1000000
+# define CUPS_SERVER_REMOTE_ANY		"_remote_any"
+#endif /* CUPS < 1.3 */
+
+#if CUPS_VERSION_MAJOR < 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR < 4)
+# define HTTP_AUTHORIZATION_CANCELED	1000
+#endif /* CUPS < 1.4 */
+
+#if CUPS_VERSION_MAJOR < 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR < 5)
+# define HTTP_PKI_ERROR			1001
+# define IPP_AUTHENTICATION_CANCELED	0x1000
+# define IPP_PKI_ERROR			0x1001
+#endif /* CUPS < 1.5 */
+
+#if HAVE_CUPS_1_6
+# define CUPS_SERVER_REMOTE_PRINTERS	"_remote_printers"
+#endif /* CUPS > 1.5 */
+
 //////////////////////
 // Worker functions //
 //////////////////////
@@ -170,7 +189,7 @@ do_password_callback (const char *prompt)
     password = NULL;
   else
   {
-    pwval = PyString_AsString (result);
+    pwval = PyBytes_AsString (result);
     password = strdup (pwval);
   }
 
@@ -188,6 +207,138 @@ do_password_callback (const char *prompt)
 }
 #endif /* !HAVE_CUPS_1_4 */
 
+#ifndef HAVE_CUPS_1_6
+int
+ippGetBoolean(ipp_attribute_t *attr,
+              int             element)
+{
+  return (attr->values[element].boolean);
+}
+
+int
+ippGetCount(ipp_attribute_t *attr)
+{
+  return (attr->num_values);
+}
+
+ipp_tag_t
+ippGetGroupTag(ipp_attribute_t *attr)
+{
+  return (attr->group_tag);
+}
+
+int
+ippGetInteger(ipp_attribute_t *attr,
+              int             element)
+{
+  return (attr->values[element].integer);
+}
+
+const char *
+ippGetName(ipp_attribute_t *attr)
+{
+  return (attr->name);
+}
+
+ipp_op_t
+ippGetOperation(ipp_t *ipp)
+{
+  return (ipp->request.op.operation_id);
+}
+
+int
+ippGetRange(ipp_attribute_t *attr,
+	    int             element,
+	    int             *uppervalue)
+{
+  if (uppervalue)
+    *uppervalue = attr->values[element].range.upper;
+
+  return (attr->values[element].range.lower);
+}
+
+int
+ippGetResolution(
+    ipp_attribute_t *attr,
+    int             element,
+    int             *yres,
+    ipp_res_t       *units)
+{
+  if (yres)
+    *yres = attr->values[element].resolution.yres;
+
+  if (units)
+    *units = attr->values[element].resolution.units;
+
+  return (attr->values[element].resolution.xres);
+}
+
+ipp_status_t
+ippGetStatusCode(ipp_t *ipp)
+{
+  return (ipp->request.status.status_code);
+}
+
+const char *
+ippGetString(ipp_attribute_t *attr,
+             int             element,
+             const char      **language)
+{
+  return (attr->values[element].string.text);
+}
+
+ipp_tag_t
+ippGetValueTag(ipp_attribute_t *attr)
+{
+  return (attr->value_tag);
+}
+
+ipp_attribute_t	*
+ippFirstAttribute(ipp_t *ipp)
+{
+  if (!ipp)
+    return (NULL);
+  return (ipp->current = ipp->attrs);
+}
+
+ipp_attribute_t *
+ippNextAttribute(ipp_t *ipp)
+{
+  if (!ipp || !ipp->current)
+    return (NULL);
+  return (ipp->current = ipp->current->next);
+}
+
+int
+ippSetInteger(ipp_t           *ipp,
+              ipp_attribute_t **attr,
+              int             element,
+              int             intvalue)
+{
+  (*attr)->values[element].integer = intvalue;
+  return (1);
+}
+
+int
+ippSetOperation(ipp_t    *ipp,
+                ipp_op_t op)
+{
+  ipp->request.op.operation_id = op;
+  return (1);
+}
+
+int
+ippSetString(ipp_t           *ipp,
+             ipp_attribute_t **attr,
+             int             element,
+             const char      *strvalue)
+{
+  (*attr)->values[element].string.text = (char *) strvalue;
+  return (1);
+}
+
+#endif
+
 //////////////////////////
 // Module-level methods //
 //////////////////////////
@@ -245,9 +396,13 @@ cups_modelSort (PyObject *self, PyObject *args)
     PyErr_SetString (PyExc_RuntimeError, "Insufficient memory");
     return NULL;
   }
-
+#if PY_MAJOR_VERSION >= 3
+  PyUnicode_AsWideChar (a, wca, size_a);
+  PyUnicode_AsWideChar (b, wcb, size_b);
+#else
   PyUnicode_AsWideChar ((PyUnicodeObject *) a, wca, size_a);
   PyUnicode_AsWideChar ((PyUnicodeObject *) b, wcb, size_b);
+#endif
   Py_DECREF (a);
   Py_DECREF (b);
   return Py_BuildValue ("i", do_model_compare (wca, wcb));
@@ -256,24 +411,34 @@ cups_modelSort (PyObject *self, PyObject *args)
 static PyObject *
 cups_setUser (PyObject *self, PyObject *args)
 {
-  const char *user;
+  PyObject *userobj;
+  char *user;
 
-  if (!PyArg_ParseTuple (args, "s", &user))
+  if (!PyArg_ParseTuple (args, "O", &userobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&user, userobj) == NULL)
     return NULL;
 
   cupsSetUser (user);
+  free (user);
   Py_RETURN_NONE;
 }
 
 static PyObject *
 cups_setServer (PyObject *self, PyObject *args)
 {
-  const char *server;
+  PyObject *serverobj;
+  char *server;
 
-  if (!PyArg_ParseTuple (args, "s", &server))
+  if (!PyArg_ParseTuple (args, "O", &serverobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&server, serverobj) == NULL)
     return NULL;
 
   cupsSetServer (server);
+  free (server);
   Py_RETURN_NONE;
 }
 
@@ -303,13 +468,13 @@ cups_setEncryption (PyObject *self, PyObject *args)
 static PyObject *
 cups_getUser (PyObject *self)
 {
-  return PyString_FromString (cupsUser ());
+  return PyUnicode_FromString (cupsUser ());
 }
 
 static PyObject *
 cups_getServer (PyObject *self)
 {
-  return PyString_FromString (cupsServer ());
+  return PyUnicode_FromString (cupsServer ());
 }
 
 static PyObject *
@@ -414,6 +579,159 @@ cups_ppdSetConformance (PyObject *self, PyObject *args)
   Py_RETURN_NONE;
 }
 
+#ifdef HAVE_CUPS_1_6
+static PyObject *
+cups_enumDests (PyObject *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *cb;
+  int flags = 0;
+  int msec = -1;
+  int type = 0;
+  int mask = 0;
+  PyObject *user_data = NULL;
+  CallbackContext context;
+  int ret;
+  static char *kwlist[] = { "cb",
+			    "flags",
+			    "msec",
+			    "type",
+			    "mask",
+			    "user_data",
+			    NULL };
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|iiiiO", kwlist,
+				    &cb,
+				    &flags,
+				    &msec,
+				    &type,
+				    &mask,
+				    &user_data))
+    return NULL;
+
+  if (!PyCallable_Check (cb)) {
+    PyErr_SetString (PyExc_TypeError, "cb must be callable");
+    return NULL;
+  }
+
+  if (!user_data)
+    user_data = Py_None;
+
+  Py_XINCREF (cb);
+  Py_XINCREF (user_data);
+  context.cb = cb;
+  context.user_data = user_data;
+  ret = cupsEnumDests (flags,
+		       msec,
+		       NULL,
+		       type,
+		       mask,
+		       cups_dest_cb,
+		       &context);
+  Py_XDECREF (cb);
+  Py_XDECREF (user_data);
+
+  if (!ret) {
+    PyErr_SetString (PyExc_RuntimeError, "cupsEnumDests failed");
+    return NULL;
+  }
+
+  Py_RETURN_NONE;
+}
+#endif /* HAVE_CUPS_1_6 */
+
+#ifdef HAVE_CUPS_1_6
+static PyObject *
+cups_connectDest (PyObject *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *destobj;
+  PyObject *cb;
+  int flags = 0;
+  int msec = -1;
+  PyObject *user_data = NULL;
+  CallbackContext context;
+  char resource[HTTP_MAX_URI];
+  http_t *conn;
+  Connection *connobj;
+  Dest *dest_o;
+  cups_dest_t dest;
+  PyObject *ret;
+  static char *kwlist[] = { "dest",
+			    "cb",
+			    "flags",
+			    "msec",
+			    "user_data",
+			    NULL };
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "OO|iiO", kwlist,
+				    &destobj,
+				    &cb,
+				    &flags,
+				    &msec,
+				    &user_data))
+    return NULL;
+
+  if (Py_TYPE(destobj) != &cups_DestType) {
+    PyErr_SetString (PyExc_TypeError, "dest must be Dest object");
+    return NULL;
+  }
+
+  if (!PyCallable_Check (cb)) {
+    PyErr_SetString (PyExc_TypeError, "cb must be callable");
+    return NULL;
+  }
+
+  if (!user_data)
+    user_data = Py_None;
+
+  Py_XINCREF (cb);
+  Py_XINCREF (user_data);
+  context.cb = cb;
+  context.user_data = user_data;
+  resource[0] = '\0';
+
+  dest_o = (Dest *) destobj;
+  dest.is_default = dest_o->is_default;
+  dest.name = dest_o->destname;
+  dest.instance = dest_o->instance;
+  dest.num_options = dest_o->num_options;
+  dest.options = malloc (dest_o->num_options * sizeof (dest.options[0]));
+  int i;
+  for (i = 0; i < dest_o->num_options; i++) {
+    dest.options[i].name = dest_o->name[i];
+    dest.options[i].value = dest_o->value[i];
+  }
+
+  conn = cupsConnectDest (&dest,
+			  flags,
+			  msec,
+			  NULL,
+			  resource,
+			  sizeof (resource),
+			  cups_dest_cb,
+			  &context);
+  Py_XDECREF (cb);
+  Py_XDECREF (user_data);
+  free (dest.options);
+
+  if (!conn) {
+    set_ipp_error (cupsLastError (), cupsLastErrorString ());
+    return NULL;
+  }
+
+  PyObject *largs = Py_BuildValue ("()");
+  PyObject *lkwlist = Py_BuildValue ("{}");
+  connobj = (Connection *) PyType_GenericNew (&cups_ConnectionType,
+					      largs, lkwlist);
+  Py_DECREF (largs);
+  Py_DECREF (lkwlist);
+  connobj->host = strdup ("");
+  connobj->http = conn;
+
+  ret = Py_BuildValue ("(Os)", (PyObject *) connobj, resource);
+  return ret;
+}
+#endif /* HAVE_CUPS_1_6 */
+
 static PyObject *
 cups_require (PyObject *self, PyObject *args)
 {
@@ -456,7 +774,19 @@ fail:
   return NULL;
 }
 
-static PyMethodDef CupsMethods[] = {
+struct module_state {
+    PyObject *error;
+};
+
+#if PY_MAJOR_VERSION >= 3
+#define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
+#else
+#define GETSTATE(m) (&_state)
+static struct module_state _state;
+#endif
+
+
+static PyMethodDef cups_methods[] = {
   { "modelSort", cups_modelSort, METH_VARARGS,
     "modelSort(s1,s2) -> integer\n\n"
     "Sort two model strings.\n\n"
@@ -536,6 +866,43 @@ static PyMethodDef CupsMethods[] = {
     "@type level: integer\n"
     "@param level: PPD_CONFORM_RELAXED or PPD_CONFORM_STRICT" },
 
+#ifdef HAVE_CUPS_1_6
+  { "enumDests",
+    (PyCFunction) cups_enumDests, METH_VARARGS | METH_KEYWORDS,
+    "enumDests(cb,flags=0,msec=-1,type=0,mask=0,user_data=None) -> None\n\n"
+    "@type cb: callable\n"
+    "@param cb: callback function, given user_data, dest flags, and dest.\n"
+    "Should return 1 to continue enumeration and 0 to cancel.\n"
+    "@type flags: integer\n"
+    "@param flags: enumeration flags\n"
+    "@type msec: integer\n"
+    "@param msec: timeout, or -1 for no timeout\n"
+    "@type type: integer\n"
+    "@param type: bitmask of printer types to return\n"
+    "@type mask: integer\n"
+    "@param mask: bitmask of type bits to examine\n"
+    "@type user_data: object\n"
+    "@param user_data: user data to pass to callback function\n"},
+#endif /* HAVE_CUPS_1_6 */
+
+#ifdef HAVE_CUPS_1_6
+  { "connectDest",
+    (PyCFunction) cups_connectDest, METH_VARARGS | METH_KEYWORDS,
+    "connectDest(dest,cb,flags=0,msec=-1,user_data=None) -> (conn, resource)\n\n"
+    "@type dest: Dest object\n"
+    "@param dest: destination to connect to\n"
+    "@type cb: callable\n"
+    "@param cb: callback function, given user_data, dest flags, and dest.\n"
+    "Should return 1 to continue enumeration and 0 to cancel.\n"
+    "@type flags: integer\n"
+    "@param flags: enumeration flags\n"
+    "@type msec: integer\n"
+    "@param msec: timeout, or -1 for no timeout\n"
+    "@type user_data: object\n"
+    "@param user_data: user data to pass to callback function\n"
+    "@return: a 2-tuple of the Connection object and the HTTP resource.\n"},
+#endif /* HAVE_CUPS_1_6 */
+
   { "require", cups_require, METH_VARARGS,
     "require(version) -> None\n\n"
     "Require pycups version.\n\n"
@@ -546,17 +913,65 @@ static PyMethodDef CupsMethods[] = {
   { NULL, NULL, 0, NULL }
 };
 
+#if PY_MAJOR_VERSION >= 3
+static int cups_traverse(PyObject *m, visitproc visit, void *arg) {
+    Py_VISIT(GETSTATE(m)->error);
+    return 0;
+}
+
+static int cups_clear(PyObject *m) {
+    Py_CLEAR(GETSTATE(m)->error);
+    return 0;
+}
+
+
+static struct PyModuleDef moduledef = {
+        PyModuleDef_HEAD_INIT,
+        "cups",
+        NULL,
+        sizeof(struct module_state),
+        cups_methods,
+        NULL,
+        cups_traverse,
+        cups_clear,
+        NULL
+};
+
+#define INITERROR return NULL
+
+PyObject *
+PyInit_cups(void)
+
+#else
+#define INITERROR return
+
 void
 initcups (void)
+#endif
 {
-  PyObject *m = Py_InitModule ("cups", CupsMethods);
+#if PY_MAJOR_VERSION >= 3
+    PyObject *m = PyModule_Create(&moduledef);
+#else
+    PyObject *m = Py_InitModule("cups", cups_methods);
+#endif
+
+  if (m == NULL)
+    INITERROR;
+  struct module_state *st = GETSTATE(m);
+
+  st->error = PyErr_NewException("cups.Error", NULL, NULL);
+  if (st->error == NULL) {
+    Py_DECREF(m);
+    INITERROR;
+  }
+
   PyObject *d = PyModule_GetDict (m);
   PyObject *obj;
 
   // Connection type
   cups_ConnectionType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_ConnectionType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Connection",
 		      (PyObject *)&cups_ConnectionType);
@@ -564,7 +979,7 @@ initcups (void)
   // PPD type
   cups_PPDType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_PPDType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "PPD",
 		      (PyObject *)&cups_PPDType);
@@ -572,7 +987,7 @@ initcups (void)
   // Option type
   cups_OptionType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_OptionType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Option",
 		      (PyObject *)&cups_OptionType);
@@ -580,7 +995,7 @@ initcups (void)
   // Group type
   cups_GroupType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_GroupType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Group",
 		      (PyObject *)&cups_GroupType);
@@ -588,7 +1003,7 @@ initcups (void)
   // Constraint type
   cups_ConstraintType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_ConstraintType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Constraint",
 		      (PyObject *)&cups_ConstraintType);
@@ -596,7 +1011,7 @@ initcups (void)
   // Attribute type
   cups_AttributeType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_AttributeType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Attribute",
 		      (PyObject *)&cups_AttributeType);
@@ -604,7 +1019,7 @@ initcups (void)
   // Dest type
   cups_DestType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_DestType) < 0)
-    return;
+    INITERROR;
 
   PyModule_AddObject (m, "Dest",
 		      (PyObject *)&cups_DestType);
@@ -612,7 +1027,7 @@ initcups (void)
   // IPPRequest type
   cups_IPPRequestType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_IPPRequestType) < 0)
-      return;
+      INITERROR;
 
   PyModule_AddObject (m, "IPPRequest",
 		      (PyObject *)&cups_IPPRequestType);
@@ -620,17 +1035,21 @@ initcups (void)
   // IPPAttribute type
   cups_IPPAttributeType.tp_new = PyType_GenericNew;
   if (PyType_Ready (&cups_IPPAttributeType) < 0)
-      return;
+      INITERROR;
 
   PyModule_AddObject (m, "IPPAttribute",
 		      (PyObject *)&cups_IPPAttributeType);
 
   // Constants
-
+#if PY_MAJOR_VERSION >= 3
+  #define INT_CONSTANT(name)					\
+    PyDict_SetItemString (d, #name, PyLong_FromLong (name))
+#else
 #define INT_CONSTANT(name)					\
   PyDict_SetItemString (d, #name, PyInt_FromLong (name))
+#endif
 #define STR_CONSTANT(name)					\
-  PyDict_SetItemString (d, #name, PyString_FromString (name))
+  PyDict_SetItemString (d, #name, PyUnicode_FromString (name))
 
   // CUPS printer types
   INT_CONSTANT (CUPS_PRINTER_LOCAL);
@@ -659,9 +1078,7 @@ initcups (void)
   INT_CONSTANT (CUPS_PRINTER_AUTHENTICATED);
   INT_CONSTANT (CUPS_PRINTER_COMMANDS);
   INT_CONSTANT (CUPS_PRINTER_OPTIONS);
-#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
   INT_CONSTANT (CUPS_PRINTER_DISCOVERED);
-#endif /* CUPS 1.3 */
 
   // HTTP encryption
   INT_CONSTANT (HTTP_ENCRYPT_IF_REQUESTED);
@@ -669,6 +1086,14 @@ initcups (void)
   INT_CONSTANT (HTTP_ENCRYPT_REQUIRED);
   INT_CONSTANT (HTTP_ENCRYPT_ALWAYS);
 
+  // Document formats
+  STR_CONSTANT (CUPS_FORMAT_AUTO);
+  STR_CONSTANT (CUPS_FORMAT_COMMAND);
+  STR_CONSTANT (CUPS_FORMAT_PDF);
+  STR_CONSTANT (CUPS_FORMAT_POSTSCRIPT);
+  STR_CONSTANT (CUPS_FORMAT_RAW);
+  STR_CONSTANT (CUPS_FORMAT_TEXT);
+
   // Selected HTTP status codes
   INT_CONSTANT (HTTP_ERROR);
   INT_CONSTANT (HTTP_OK);
@@ -685,12 +1110,8 @@ initcups (void)
   INT_CONSTANT (HTTP_SERVICE_UNAVAILABLE);
   INT_CONSTANT (HTTP_GATEWAY_TIMEOUT);
   INT_CONSTANT (HTTP_NOT_SUPPORTED);
-#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 4)
   INT_CONSTANT (HTTP_AUTHORIZATION_CANCELED);
-#endif /* CUPS 1.4 */
-#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 5)
   INT_CONSTANT (HTTP_PKI_ERROR);
-#endif /* CUPS 1.5 */
 
   // PPD UI enum
   INT_CONSTANT (PPD_UI_BOOLEAN);
@@ -810,10 +1231,8 @@ initcups (void)
   INT_CONSTANT (IPP_ERROR_JOB_CANCELLED);
   INT_CONSTANT (IPP_MULTIPLE_JOBS_NOT_SUPPORTED);
   INT_CONSTANT (IPP_PRINTER_IS_DEACTIVATED);
-#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 5)
   INT_CONSTANT (IPP_AUTHENTICATION_CANCELED);
   INT_CONSTANT (IPP_PKI_ERROR);
-#endif /* CUPS 1.5 */
 
   // IPP states
   INT_CONSTANT (IPP_ERROR);
@@ -853,44 +1272,58 @@ initcups (void)
   STR_CONSTANT (CUPS_SERVER_REMOTE_PRINTERS);
   STR_CONSTANT (CUPS_SERVER_SHARE_PRINTERS);
   STR_CONSTANT (CUPS_SERVER_USER_CANCEL_ANY);
-#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 3)
   STR_CONSTANT (CUPS_SERVER_REMOTE_ANY);
-#endif /* CUPS 1.3 */
+
+#ifdef HAVE_CUPS_1_6
+  // Dest enumeration flags
+  INT_CONSTANT (CUPS_DEST_FLAGS_NONE);
+  INT_CONSTANT (CUPS_DEST_FLAGS_UNCONNECTED);
+  INT_CONSTANT (CUPS_DEST_FLAGS_MORE);
+  INT_CONSTANT (CUPS_DEST_FLAGS_REMOVED);
+  INT_CONSTANT (CUPS_DEST_FLAGS_ERROR);
+  INT_CONSTANT (CUPS_DEST_FLAGS_RESOLVING);
+  INT_CONSTANT (CUPS_DEST_FLAGS_CONNECTING);
+  INT_CONSTANT (CUPS_DEST_FLAGS_CANCELED);
+#endif /* HAVE_CUPS_1_6 */
 
   // Exceptions
   obj = PyDict_New ();
-  PyDict_SetItemString (obj, "__doc__", PyString_FromString(
+  PyDict_SetItemString (obj, "__doc__", PyUnicode_FromString(
     "This exception is raised when an HTTP problem has occurred.  It \n"
     "provides an integer HTTP status code.\n\n"
     "Use it like this::\n"
     "  try:\n"
     "    ...\n"
-    "  except cups.HTTPError (status):\n"
+    "  except cups.HTTPError as (status):\n"
     "    print 'HTTP status is %d' % status\n"));
   HTTPError = PyErr_NewException ("cups.HTTPError", NULL, obj);
   Py_DECREF (obj);
   if (HTTPError == NULL)
-    return;
+    INITERROR;
   Py_INCREF (HTTPError);
   PyModule_AddObject (m, "HTTPError", HTTPError);
 
   obj = PyDict_New ();
-  PyDict_SetItemString (obj, "__doc__", PyString_FromString(
+  PyDict_SetItemString (obj, "__doc__", PyUnicode_FromString(
     "This exception is raised when an IPP error has occurred.  It \n"
     "provides an integer IPP status code, and a human-readable string \n"
     "describing the error.\n\n"
     "Use it like this::\n"
     "  try:\n"
     "    ...\n"
-    "  except cups.IPPError (status, description):\n"
+    "  except cups.IPPError as (status, description):\n"
     "    print 'IPP status is %d' % status\n"
     "    print 'Meaning:', description\n"));
   IPPError = PyErr_NewException ("cups.IPPError", NULL, obj);
   Py_DECREF (obj);
   if (IPPError == NULL)
-    return;
+    INITERROR;
   Py_INCREF (IPPError);
   PyModule_AddObject (m, "IPPError", IPPError);
+
+#if PY_MAJOR_VERSION >= 3
+    return m;
+#endif
 }
 
 ///////////////
